"""
An example of a small application which control a single GPIO pin.

Overview
--------

This examples export a single noun, `led`, which is used to set, test and otherwise control a single GPIO pin. It can be used an example of a the way in which this API class can be used, along with the definition of the  `urest.examples.simpleled.SimpleLED` class.

To use this example, the variables `SSID` and `PASSWORD` must also be defined, and set to suitable values for the local network. Space for these variables has been provided at the start of the '`main`' loop. The code will then attempt to create an API server on the _network_ side of the machines local interfaces, bound to port 80.

.. Note:: MicroPython Includes

    This module as presented has a number of `try..finally` blocks which exist to
    catch uses of this code where the standard MicroPython libraries are not
    available. If _known_ to be running under MicroPython, or suitably adapted,
    these can be removed.

.. Warning:: Exposed Network Ports

    This code _does not_ attempt to take any defensive stances with respect
    to the network environment, beyond the normal measures in the API code.
    As is normal for MicroPython implementations, it also assumes the SSID
    and the network password are embedded in the code, and therefore exposed.
    Careful review of the application code, and the network environment, should
    be undertaken before this code is used in anything other than a benign
    environment under _total_ control of the developer.

Tested Implementations
----------------------

This version is written for MicroPython 3.4, and has been tested on:

  * Raspberry Pi Pico W

Licence
-------

This code, and all included code, is made available under the terms of the MIT Licence

> Copyright 2021--2022 (c) Erik de Lange, Copyright (c) 2022--2023 David Love

> Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

> The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

> THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import time

try:
    from machine import Pin
    from micropython import const

    import network
    import uasyncio as asyncio
except ImportError:
    print("Ignoring MicroPython includes")

from urest.http import RESTServer
from urest.api import APIBase
from urest.examples.simpleled import SimpleLED

###
### Main Loop
###

# Set the SSID of the wireless network
SSID = "SSID"
# Set the password of the wireless network
PASSWORD = "PASSWORD"

# Set-up the link status LED
try:
    link_status = Pin("WL_GPIO0", Pin.OUT)
except NameError:
    print("Cannot find the MicroPython PIN library")

try:
    # Set-up the Wireless Driver
    wlan = network.WLAN(network.STA_IF)
    wlan.active(True)
    wlan.connect(SSID, PASSWORD)

    # Number of connection attempts
    # before a hard fail
    max_wait = 10

    while max_wait > 0:
        if wlan.status() < 0 or wlan.status() >= 3:
            print("Already connected!")
            print("IP:  " + wlan.ifconfig()[0])
            break

        max_wait -= 1
        print("waiting for connection...")
        time.sleep(1)

        print(wlan.status())

        # Handle connection error
        if wlan.status() != 3:
            raise RuntimeError("network connection failed")
        else:
            print("Connected")
            print("IP: " + wlan.ifconfig()[0])

    # Display the link light if connected
    if wlan.status() == 3:
        link_status.on()
    else:
        link_status.off()

except NameError:
    print("Cannot find the MicroPython network library")

# Create the server ...
app = RESTServer()

# ... and register the nouns
try:
    app.register_noun("green_led0", SimpleLED(1))
except NameError:
    print("Cannot find the MicroPython PIN library")

if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    loop.create_task(app.start())
    loop.run_forever()



"""
Micro HTTP server dedicated to REST-style API Requests. Inspired by the
[MicroPython HTTP Server](https://github.com/erikdelange/MicroPython-HTTP-Server)
by Erik de Lange, as a simple consumer of the Python 3 `asyncio` library for the
low-level socket handling and use of co-routines to simplify request handling.

This version is written for MicroPython 3.4, and has been tested on:

  * Raspberry Pi Pico W


Standards
---------

  * For the HTTP/1.1 specification see: https://www.ietf.org/rfc/rfc2616.txt
  * For the JSON specification see: https://www.ecma-international.org/publications-and-standards/standards/ecma-404

Licence
-------

This module, and all included code, is made available under the terms of the MIT
Licence

> Copyright 2022 (c) Erik de Lange, Copyright (c) 2022-2023 David Love

> Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

> The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

> THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

# Import the standard error library
import errno

# Import the Asynchronous IO Library, preferring the MicroPython library if
# available
try:
    import uasyncio as asyncio
except ImportError:
    import asyncio

# Import the RAWResponse class
from .response import HTTPResponse

# Import the API Base class
from ..api.base import APIBase

##
## Constants
##

ASCII_UPPERCASE = set("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ")
""" Constant for the set of ASCII letters """
ASCII_DIGITS = set("0123456789")
""" Constant for the set of ASCII digits """
ASCII_EXTRA = set("_")
""" Constant for the extra ASCII characters allowed in the URI """

JSON_TYPE_INT = 0
""" Constant for JSON token type of Integer """
JSON_TYPE_STR = 1
""" Constant for JSON token type of String """
JSON_TYPE_ERROR = 1
""" Constant for JSON token type of error/unknown """

##
## Exceptions
##


class RESTServerError(Exception):
    pass


class RESTParseError(Exception):
    pass


##
## Classes
##


class RESTServer:

    #
    def __init__(
        self,
        host: str = "0.0.0.0",
        port: int = 80,
        backlog: int = 5,
        read_timeout: int = 30,
        write_timeout: int = 5,
    ):
        """Initialise the server with reasonable defaults. These should work
        for most cases, and should be set so that most clients won't have
        to touch them.

        The `urest.http.server.RESTServer` class acts as the primary interface
        to the library, handling all the network communication with the client,
        formatting the response and marshalling the API calls required to generate that
        response.

        In most cases consumers of this module will create a single instance of
        the `urest.http.server.RESTServer` class, and then pass the reference to the
        `urest.http.server.RESTServer.start` method to an event loop of the `asyncio`
        library.

        For example the following code creates a variable `app` for the instance
        of the `urest.http.server.RESTServer` class, and passes this to the 'main' event
        loop of the `asyncio` library

        ```python
        app = RESTServer()


        if __name__ == "__main__":
          loop = asyncio.get_event_loop()
          loop.create_task(app.start())
          loop.run_forever()
        ```

        The `urest.http.server.RESTServer.start` method is expected to be used
        in a [`asyncio` event
        loop](https://docs.python.org/3.4/library/asyncio-eventloop.html), as above;
        with the tasks being handled by the `urest.http.server.RESTServer.dispatch_noun`
        method. If the event loop is required to be closed, or destroyed, the tasks
        can be removed using the `urest.http.server.RESTServer.stop` method.

        .. Note::
          The code in this class assumes the `asyncio` library with an interface
          roughly equivalent to Python 3.4: although the MicroPython module supports _some_
          later extensions. Given the code churn in the `asyncio` module between Python 3.4
          and Python 3.10, careful testing is required to ensure implementation compatibility.

        Parameters
        ----------

        host: string
            A resolvable DNS host name or IP address. Note that the exact requirements are
            determined by the
            [`asyncio.BaseEventLoop.create_server`](https://docs.python.org/3.4/library/
            asyncio-eventloop.html#asyncio.BaseEventLoop.create_server) method, which should
            be checked carefully for implementation defined limitations.

            **Default:** An IPv4 sock on the local host.

        port: integer
            The local (server) port to bind the socket to. Note that the exact requirements are
            determined by the
            [`asyncio.BaseEventLoop.create_server`](https://docs.python.org/3.4/library/
            asyncio-eventloop.html#asyncio.BaseEventLoop.create_server) method, which should
            be checked carefully for implementation defined limitations (e.g. extra privileges
            required for system ports).

            **Default:** The IANA Assigned port 80 for an HTTP Server.

        backlog: integer
            Roughly the size of the pool of connections for the underlying
            `socket`. Once this value has been exceeded, the tasks will be suspended by the
            co-routine handler until the underlying `socket` can clear them. Note that the
            size (and interpretation) of this value is system dependent: see the [`socket`
            API](https://docs.python.org/3.4/library/socket.html#module-socket) for more
            details.

            **Default:** 5 (typically the maximum pool size allowed).

        read_timeout: integer
            Length of time in seconds to wait for a response from the client before declaring
            failure.

            **Default:** 30 seconds.

        write_timeout: integer
            Length of time in seconds to wait for the network socket to accept a write to the
            client, before declaring failure.

            **Default:** 5 seconds.

        """

        self.host = host
        self.port = port
        self.backlog = backlog
        self.read_timeout = read_timeout
        self.write_timeout = write_timeout
        self._server = None
        self._nouns = {}

    def parse_data(self, data_str: str) -> dict:
        """
        Attempt to parse a string containing JSON-like formatting into a single dictionary.

        This function is **very** far from a full JSON parser: quite deliberately it will
        only accept a single object of name/value pairs. Any arrays are **not** accepted.
        In addition, this parser will coerce the 'name' side of the dictionary into a
        string; or if this cannot be done will raise a 'RESTParseError'. The 'value' of a
        name/value pair will like-wise be coerced into its JSON type; or a 'RESTParseError'
        raised if this cannot be done.

        The parsing is done via a very simple stack-based parser, assuming no backtracking.
        This will cope with valid JSON: but will quickly abort if the JSON is malformed,
        raising a 'RESTParseError'. This is quite deliberate as we will only accept valid
        JSON from the client: if we can't parse the result that is the clients problem...

        The parser will also finish after the first found JSON object. We are expecting
        only a single dictionary from the client, and so attempts to add something more
        exotic will be ignored.

        Parameters
        ----------

        data_str: str
            A JSON object string, representing a single dictionary

        Returns
        -------

        dict
          A mapping of (key, value) pairs which defines the dictionary of the `data_str`
          object. All `key` values will be in Python string format: values will be as
          defined in the `data_str` object.
        """

        return_dictionary = {}
        parse_stack = []
        object_start = False

        token_start = False
        token_sep = False

        token_type = JSON_TYPE_INT
        token_str = ""

        for char in data_str.decode("ascii"):

            # Look for the first object
            if char in ["{"]:
                object_start = True

            # If we are inside an object, attempt to assemble the
            # tokens
            if object_start:

                # If a '"' is found...
                if char in ['"']:

                    # ... and if we are building a token, this should be the
                    # end of a string, so push it to the stack ...
                    if token_start:
                        if token_type == JSON_TYPE_STR:
                            parse_stack.append(token_str)

                            token_start = False
                        else:
                            raise RESTParseError("Invalid string termination")

                        token_type = JSON_TYPE_INT
                        token_sep = False
                        token_start = False
                        token_str = ""

                    # ... Otherwise if we are not building a token, set the
                    # type marker, and start building a new string
                    else:
                        token_type = JSON_TYPE_STR
                        token_start = True
                        token_str = ""

                # Look for the separator between a 'key' and a 'value'
                if char in [":"]:
                    token_sep = True

                # Look for the end of a token
                if char in [",", "}"]:

                    # Add the key/value to the return dictionary if
                    # it appears to be valid
                    if token_sep:

                        if token_type == JSON_TYPE_INT:
                            # The token won't have been terminated
                            # so still should be in `token_str`
                            #
                            # NOTE: Technically an unterminated String
                            #       is an error, so will parsing will
                            #       break here (and we won't care)
                            parse_stack.append(token_str)

                        value = parse_stack.pop()
                        key = parse_stack.pop()

                        if token_type == JSON_TYPE_STR:
                            return_dictionary[key] = str(value)

                        if token_type == JSON_TYPE_INT:
                            return_dictionary[key] = int(value)

                    # If this is the end of the object, then return ...
                    if char in ["}"]:
                        return return_dictionary

                    # .. otherwise, cleanup and continue
                    else:
                        token_type = JSON_TYPE_ERROR
                        token_sep = False
                        token_start = False

                # If this isn't anything interesting, assume it is part of a token
                if object_start and (
                    (char in ASCII_UPPERCASE) or (char in ASCII_DIGITS)
                ):
                    token_str = token_str + char

        return return_dictionary

    def register_noun(self, noun: str, handler: APIBase):
        """
        Register a new object handler for the noun passed by the client.

        Parameters
        ----------

        noun: string
            String representing the noun to use in the API

        handler: `urest.api.base.APIBase`
            Object handling the request from the client
        """

        old_handler = None
        print("Register!")

        try:

            if noun in self._nouns:
                old_handler = self._nouns[noun]

            if isinstance(noun, str) and isinstance(handler, APIBase):
                self._nouns[noun.lower()] = handler

        except:
            if old_handler is not None:
                self._nouns[noun] = old_handler

    async def dispatch_noun(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ):
        """
        Core client handling routine. This connects the `reader` and `writer`
        streams from the IO library to the API requests detailed by the rest
        of the server. Most of the work is done elsewhere, by the API handlers:
        this is mostly a sanity check and a routing engine.

        .. Warning::
          This routine _must_ handle arbitrary network traffic, and so
          **must** be as defensive as possible to avoid security issues in
          the API layer which results from arbitrary input stuffing and
          alike. Assume that anything from the `reader` is potentially
          dangerous to the health of the API layer: unless shown otherwise...



        Parameters
        ----------

        reader: `asyncio.StreamReader`
            An asynchronous stream, representing the network response _from_ the
            client. This is usually set-up indirectly by the `asyncio` library as part of a
            network response to the client, and will be represented by an
            [asyncio.StreamReader](https://docs.python.org/3.4/library/asyncio-stream.html#
            streamreader).

        writer: `asyncio.StreamWriter`
            An asynchronous stream, representing the network response _to_ the
            client. This is usually set-up indirectly by the `asyncio` library as part of a
            network response to the client, and will be represented by an
            [asyncio.StreamWriter](https://docs.python.org/3.4/library/asyncio-stream.html#
            streamwriter).
        """

        # Attempt the parse whatever rubbish the client sends, and assemble the
        # fragments into an API request. Any failures should result in an
        # `Exception`: success should result in an API call
        try:

            # Get the raw network request and decode into UTF-8
            request_uri = await asyncio.wait_for(reader.readline(), self.read_timeout)
            request_uri = request_uri.decode("utf8")

            # Check for empty requests, and if found terminate the connection
            if request_uri in [b"", b"\r\n"]:
                # DEBUG
                print(
                    f"CLIENT: [{writer.get_extra_info('peername')[0]}] Empty request line"
                )
                return

            # DEBUG
            print(
                f"CLIENT URI : [{writer.get_extra_info('peername')[0]}] {request_uri}"
            )

            # Get the header of the request, if it is available, decoded into UTF-8
            request_header = {}
            request_line = None

            while not request_line in [b"", b"\r\n"]:
                request_line = await asyncio.wait_for(
                    reader.readline(), self.read_timeout
                )

                if request_line.find(b":") != -1:
                    name, value = request_line.split(b":", 1)
                    request_header[name.decode("utf-8").lower()] = value.decode(
                        "utf-8"
                    ).strip()

            # DEBUG
            print(
                f"CLIENT HEAD: [{writer.get_extra_info('peername')[0]}] {request_header}"
            )

            # Check if there is a body to follow the header ...
            request_body = {}

            if "content-length" in request_header:
                request_length = int(request_header["content-length"])
                print(f"content_length: {request_length}")

                # ... check if there is _really a body to follow ...
                if request_length > 0:

                    # ... if so, get the rest of the body of the request, decoded into UTF-8

                    try:
                        request_length = int(request_header["content-length"])
                        request_data = await asyncio.wait_for(
                            reader.read(request_length), self.read_timeout
                        )
                        print(request_data)
                        request_body = self.parse_data(request_data)
                    except Exception as e:
                        print(e)
                        request_body = {}

                # DEBUG
                print(
                    f"CLIENT BODY: [{writer.get_extra_info('peername')[0]}] {request_body}"
                )

            else:
                # DEBUG
                print("CLIENT BODY: NONE")

            ## NOTE: Below is a somewhat long-winded approach to working out
            ##       the verb is based on the longest assumed verb:
            ##       '`DELETE`'. To avoid later parsing errors, and to
            ##       filter out the rubbish which might cause security
            ##       issues, we will search first for a 'space' within
            ##       the first six characters; then take either the first
            ##       six characters or the string up to the 'space'
            ##       whichever is shorter. These can then be compared
            ##       for sanity before we run the dispatcher

            # Work out the action we need to take ...

            first_space = request_uri.find(" ", 0, 7)

            if first_space > 7:
                first_space = 7

            verb = request_uri[0:first_space].upper()

            # ... Work out the noun defining the class we need to use to resolve the
            # action ...

            uri_root = request_uri.find("/", first_space)

            noun = ""
            start_noun = False

            for char in request_uri[uri_root:]:
                if (
                    (char in ASCII_UPPERCASE)
                    or (char in ASCII_DIGITS)
                    or (char in ASCII_EXTRA)
                ):
                    start_noun = True
                    noun = noun + char
                else:
                    if start_noun:
                        break

            # ... and then call the appropriate handler
            response = HTTPResponse()

            if verb == "DELETE":
                self._nouns[noun.lower()].delete_state()
                response.body = ""

            elif verb == "GET":
                return_state = self._nouns[noun.lower()].get_state()
                response_str = "{"

                try:
                    for key in return_state:
                        if isinstance(return_state[key], int):
                            response_str = (
                                response_str + f'"{key.lower()}": {return_state[key]},'
                            )
                        else:
                            response_str = (
                                response_str
                                + f'"{key.lower()}": "{return_state[key]}",'
                            )
                finally:
                    # Properly terminate the body
                    response_str = response_str[:-1] + "}"

                response.body = response_str

            elif verb == "POST":
                self._nouns[noun.lower()].set_state(request_body)
                response.body = ""

            elif verb == "PUT":
                self._nouns[noun.lower()].set_state(request_body)
                response.body = ""

            else:
                # Clearly not one of ours
                response.body = "<http><body><p>Invalid Request</p></body></http>"
                response.status = "NOT_OK"

            await response.send(writer)

            writer.write("\n")

            await writer.drain()

        # Deal with any exceptions. These are mostly client errors, and since the REST
        # API _should_ be idempotent, the client _should_ be able to simply retry. So
        # we won't do anything very fancy here
        except asyncio.TimeoutError:
            pass
        except Exception as e:
            if e.args[0] == errno.ECONNRESET:  # connection reset by client
                pass
            else:
                raise e

            print(f"!EXCEPTION!: {e}")

            response.body = "<http><body><p>Invalid Request</p></body></http>"
            response.status = "NOT_OK"

        # In principle the response should have been sent back to the client by now.
        # But we will give it one last try, and then also try to close the connection
        # cleanly for the client. This may not work due to the earlier exceptions: but
        # we will try anyway
        finally:
            # Do a soft close, dropping our end of the connection
            # to see if the client closes ...
            await writer.drain()
            writer.close()

            # ... if the client doesn't take the hint, wait
            # for `write_timeout` seconds and then force the close
            await asyncio.sleep(self.write_timeout)
            await writer.wait_closed()

    async def start(self):
        """
        Attach the method `urest.http.server.RESTServer.dispatch_noun` to an `asyncio`
        event loop, allowing the `urest.http.server.RESTServer.dispatch_noun` method to
        handle tasks representing network events from the client.

        Most of the implementation of this method is handled by
        [asyncio.start_server](https://docs.python.org/3.4/library/asyncio-stream.html#
        asyncio.start_server). In particular the
        [asyncio.start_server](https://docs.python.org/3.4/library/asyncio-stream.html#
        asyncio.start_server) method is responsible for setting up the lower-level
        networks socket, using the `host` and `port` class attributes holding the server
        (local) elements of the TCP/IP tuple. Lower level timers and queues are also
        provided by the resolution of the
        [asyncio.start_server](https://docs.python.org/3.4/library/asyncio-stream.html#
        asyncio.start_server) method, with the class attribute `backlog` being used to
        set the client (downstream) timeout.
        """

        print(f"SERVER: Started on {self.host}:{self.port}")
        self._server = await asyncio.start_server(
            self.dispatch_noun, self.host, self.port, self.backlog
        )

    async def stop(self):
        """
        Remove the tasks from an event loop, in preparation for the termination
        of that loop. Most of the implementation of this method is handled by the
        [`close`
        method](https://docs.python.org/3.4/library/asyncio-protocol.html#asyncio.
        BaseTransport.close) of `asyncio.BaseTransport`.
        """

        if self._server is not None:
            self._server.close()
            await self._server.wait_closed()
            self._server = None
            print("SERVER: Stopped")
        else:
            print("SERVER: Not started")



"""
An example of a 'noun' class, able to serve as the basis for control of an LED attached to a GPIO pin

Overview
--------

This is a reasonably minimal example of a 'noun' class, which inherits from
`urest.api.base.APIBase`. It also requires the `Pin` library from MicroPython:
but should be able to be adapted to other GPIO libraries which provide a similar
interface.

For an example of an application which uses this library, see: `urest.examples.led_control`

Tested Implementations
----------------------

This version is written for MicroPython 3.4, and has been tested on:

  * Raspberry Pi Pico W

Licence
-------

This class, and all included code, is made available under the terms of the MIT Licence

> Copyright 2021--2022 (c) Erik de Lange, Copyright (c) 2022--2023 David Love

> Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

> The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

> THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


# Import the Asynchronous IO Library, preferring the MicroPython library if
# available
try:
    from machine import Pin
except ImportError:
    print("Ignoring MicroPython include: machine")

from ..api.base import APIBase


class SimpleLED(APIBase):
    def __init__(self, pin):
        self._gpio = Pin(pin, Pin.OUT)
        self._gpio.off()

        self._state_attributes = dict(led=0)

    def set_state(self, state_attributes: dict):
        try:
            self._state_attributes["led"] = state_attributes["led"]

            if self._state_attributes["led"] == 0:
                self._gpio.off()
            else:
                self._gpio.on()

        except:
            # On exception try to return to a known good
            # state
            self._gpio.off()
            self._state_attributes["led"] = 0

    def get_state(self) -> dict:
        return dict(led=self._gpio.value())

    def delete_state(self):
        self._gpio.off()
        self._state_attributes["led"] = 0

    def update_state(self, state_attributes: dict):
        if self._state_attributes["led"] == 0:
            self._gpio.on()
            self._state_attributes["led"] = 1
        else:
            self._gpio.off()
            self._state_attributes["led"] = 0



"""
Abstract base class and helper classes for the server-provided REST API

Overview
--------

During the marshalling of client requests (provided by the
`urest.http.server.RESTServer` class), the path portion of the URI is extracted
from the client. This path is expected to take the form of

> `/< noun >`

with the 'verb' denoted by the HTTP methods. Each 'noun' is implemented by a
class which inherits from `urest.api.base.APIBase`.

.. Note::

  All nouns are checked by`urest.http.server.RESTServer`, and nouns will not be
  properly routed unless `urest.api.base.APIBase` is an ancestor of the derived
  class.

The 'noun' determines the resource of the request, and the 'verb' the action to
be taken on that resource. Also important is the intent of the client as
inferred from the underlying HTTP request, e.g. `GET` or `PUT`. Taken together
the URI and the client intent are mapped to methods of `urest.api.base.APIBase`
as follows

| Verb   | HTTP Method | `urest.api.base.APIBase` method      |
|--------|-------------|--------------------------------------------|
| Get    | `GET`       | `urest.api.base.APIBase.get_state`    |
| Set    | `PUT`       | `urest.api.base.APIBase.set_state`    |
| Update | `POST`      | `urest.api.base.APIBase.update_state` |
| Delete | `DELETE`    | `urest.api.base.APIBase.delete_state` |


Example
--------------

### Getting the Noun State

See `urest.examples.simpleled.SimpleLED` class a minimal implementation of a noun
controlling on GPIO pin, using the MicroPython `Pin` library. For the
`urest.examples.simpleled.SimpleLED` class, the exact pin being controlled is set
during the object instantiation, via the class constructor.

Assuming the `urest.http.server.RESTServer` has been created as

```python app = RESTServer() ```

then a 'noun' `led` can be registered via the
`urest.http.server.RESTServer.register_noun` method as

```python app.register_noun('led', SimpleLED(28)) ```

This will also bind the controlled GPIO pin to `Pin 28`: documentation for the
micro-controller and platform being used will be need to determine suitable
values.

Once the 'noun' has been registered, then the state of GPIO `Pin 28` can be
found by the HTTP request

```
GET /led HTTP 1.1
```

### All Lookup's Are Case Insensitive

All nouns are registered with `urest.http.server.RESTServer` in **lowercase**.
Likewise all routing of API requests will also assume that the name of the noun
to be used will be in lowercase. Thus, for instance, calling

```python app.register_noun('leD', SimpleLED(28)) ```

or

```python app.register_noun('LED', SimpleLED(28)) ```

will **all** route to the name noun with the canonical name `led`. Likewise
the request

``` GET /lEd HTTP 1.1 ```

or

``` GET /LED HTTP 1.1 ```

will **also** route to the same noun with the canonical name `led`.

### Setting the Noun State

Setting the state of the noun is a little more involved, as this will require
the desired state of the noun to be sent to the server. A useful tool for
testing purposes is the [curl](https://curl.se) utility; available on most
platforms.

Continuing the minimal example above, the command line

```
curl -X PUT -d '{"led":"0"}' -H "Content-Type: application/json" http://10.0.30.225/LED
```

will transmit something like the following HTTP request to
`urest.http.server.RESTServer`

```
PUT /LED HTTP/1.1

Host: 10.0.30.225
User-Agent: curl/7.81.0
Accept: */*
Content-Type: application/json
Content-Length: 11

{"led":"0"}
```

.. NOTE:: JSON is Required

  The only format accepted by the `urest.http.server.RESTServer` for the state
  of the nouns is [JSON](https://www.json.org/json-en.html). The
  `urest.http.server.RESTServer` will also only accept a sub-set of the JSON
  standard: notably assuming a single object, and a collection of key/value pairs.

The exact interpretation of the JSON is left to the implementation of the noun,
but in most cases the name of the noun is used to set the state value. For
instance in the above example the key `led` is interpreted as referring to the
noun, and the value `0` as the state value. In this case setting the GPIO `Pin
28` to the value `0` (or `off`). A close correspondence between the name of the
noun as referred to the API, and the name of the noun in the state list is
**strongly** advised: but is not strictly required.

Tested Implementations
----------------------

This version is written for MicroPython 3.4, and has been tested on:

  * Raspberry Pi Pico W

Licence
-------

This module, and all included code, is made available under the terms of the MIT Licence

> Copyright (c) 2022--2023 David Love

> Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

> The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

> THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

### Expose the `http` module interface
from .base import APIBase



"""
Implements the 'Abstract' Base Class for all the nouns used by the
`urest.http.server.RESTServer` class in defining resources.


.. Note::
  MicroPython does not implement the
  [`abc` module](https://docs.python.org/3.10/library/abc.html) which provides
  language-level support for abstract base classes. Therefore the
  `urest.http.server.RESTServer` class checks the ancestors of the class
  passed as a resource for the API to ensure they derive from `urest.api.base.APIBase`.
  Any class which _does not_ have `urest.api.base.APIBase` as an ancestor will
  therefore **not** work as a valid resource.

  This also means that `urest.api.base.APIBase` is not a pure virtual ABC, and
  so implements a minimum set of methods which hold and manipulate resource state.
  However all methods such as `urest.api.base.APIBase.get_state` and
  `urest.api.base.APIBase.set_state` are expected to be overridden in working
  implementations.

Licence
-------

This module, and all included code, is made available under the terms of the MIT
Licence

> Copyright (c) 2022-2023 David Love

> Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

> The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

> THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


class APIBase:

    ##
    ## Constructor
    ##

    def __init__(self):
        """The Abstract Base Class for the nouns, used in the response from the server.

        This base class defines the minimum interface used in marshalling requests from
        the clients by the `urest.http.server.RESTServer` class. The API defined by the
        server consists of 'nouns' representing the resources defined by the sub-classes
        of this base class, and the methods which act upon those resources (which map to
        the 'verbs' of the HTTP requests made to the `urest.http.server.RESTServer` class.

        Where data is returned to the client by the, e.g by the `get_state` method, JSON
        will ultimately be used as the encoding format. Sub-classes do not need to
        implement the saving of the object state to the client: however to assist they
        should be aware of the type of the each `value` returned as part of a `key/value`
        pair in the `Dictionary`.

        When returning data to the client, the `value` of each entry in the `Dictionary`
        will attempt to be inferred using the normal Python type library. If the type can be
        identified, then an appropriate JSON type will be used: otherwise the `value` of that
        `Dictionary` entry will be returned as a string.

        .. Warning:: Data will be returned to the client 'as is'
            No further checking on the validity (or otherwise) of the content of
            the `Dictionary` will be undertaken past this point. Anything that appear to be in
            a valid `Dictionary` will be returned to the client. It is the module consumers
            responsibility to ensure the returned data follows the form expected by those clients.
        """

        self._state_attributes = []
        """ Holds the current state and attributes of the resource """

    ##
    ## State Manipulation Methods
    ##

    def get_state(self) -> dict:
        """Returns the state of the resource.

        Return the state of the resource, as defined by the sub-classes. By default
        this method will return the contents of the private `state_attributes` `Dictionary`
        to the client; assuming that `Dictionary` has been appropriately completed
        in the processing of the resource.

        Returns
        -------

        Dictionary
          A mapping of (key, value) pairs which defines the resource state to return to the
          client.
        """

        return self._state_attributes

    def set_state(self, state_attributes: dict):
        """Sets the full state of the resource.

        The exact mechanism for updating the internal state of the resource represented
        by sub-classes is implementation defined. By default this method expects the
        _full_ state to be represented by the `state_attributes` parameters; and by
        default the new state will be exactly as stated by the `Dictionary` passed in
        through `state_attributes`.

        Parameters
        ----------

        state_attributes: Dictionary
            A list of (key, value) pairs representing the _full_ state of the
            resource. No merging of state is undertaken, or attempted.
        """

        if state_attributes is not None and isinstance(state_attributes, dict):
            self._state_attributes = state_attributes
        else:
            self._state_attributes = []

    def update_state(self, state_attributes: dict):
        """Updates the state of the resource, using the 'key/value' pairs of the
        `Dictionary` in `state_attributes`.


        The exact mechanism for updating the internal state of the resource represented
        by sub-classes is implementation defined. By default this expects to update
        only part of the state through the use of the partial state defined in
        `state_attributes`. This defines a `Dictionary` of (key, value) pairs which
        can be interpreted by sub-classes to update the true internal state as needed.

        Parameters
        ----------

        state_attributes: Dictionary
            A list of (key, value) pairs representing the _partial_ state of the
            resource. The exact mechanism for merging this partial state if left
            to the implementation of the sub-classes.
        """

        if state_attributes is not None and isinstance(state_attributes, dict):
            for key in state_attributes:
                try:
                    self._state_attributes[key] = state_attributes[key]
                except:
                    self._state_attributes[key] = None
        else:
            self._state_attributes = []

    def delete_state(self):
        """Removes the internal state of the resource, essentially 'resetting' or
        re-initialising the object.

        The exact mechanism for returning the state to the defaults are left to
        the implementation. However it is expected that once this call completes
        the internal state will be _identical_ to that of the default constructor.
        """

        self._state_attributes = []



"""
Network interface and helper classes for abstracting the underlying HTTP stream

Overview
--------

This module provides the abstraction of the underlying socket handling code,
using in marshaling requests from clients and building the response. All the
socket handling is built around the Python 3 `asyncio` library for the low-level
network interface, and for the use of co-routines to simplify request handling.

In most cases, consumers of this module only need to provide an `asynio` event
loop: see "Creating the Network Server" in the Section _Using the Module_ below.
By default the module will also bind to the _network_ address of the host on the
standard port 80 used by HTTP requests. This can be changed in the instantiation
of the  `urest.http.server.RESTServer` class.

Implementation
--------------

During the instantiation of the `urest.http.server.RESTServer` class, the resulting
object creates a socket bound to the specified address and port (or the host network
address and port 80 by default). This binding is undertaken as part of the `asyncio`


Using the Module
----------------

### Creating the Network Server

All network clients are assumed to be handled by an instance of the
`urest.http.server.RESTServer` class from the `urest.http.server` module. The
`urest.http.server.RESTServer` class provides both an event loop for the
`asyncio` library, and also takes care of the lower-level networking interface
for the clients. Unless working with multiple instances (e.g. via a thread
library), most model consumers are assumed have a single instance of the
`urest.http.server.RESTServer` --- **but the module will not check this**.

In most cases, something similar to the following will suffice

```python
  # Import the Asynchronous IO Library
  try:
    import uasyncio as asyncio
  except ImportError:
    import asyncio

  from urest.http import RESTServer

  app = RESTServer()

  if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    loop.create_task(app.start())
    loop.run_forever()
```

.. Warning::

  There is no guarantee of thread-safety in this module, and all
  'concurrency' is assumed to be via co-routines provided by the `asyncio`
  library. For an overview of how multiple clients can be handled using
  co-routines, [Async IO in Python](https://realpython.com/async-io-python) is
  highly recommended reading.

### Creating the API Responses

Most module consumers will not use the `urest.http.response` module directly:
but will instead sub-class `urest.api.base.APIBase` to provide the core of the
response to the network clients. For details of how the `urest.http.server.RESTServer` and `urest.api.base.APIBase` classes interact, the module documentation for `urest.api` should be consulted. In addition, the documentation for the `urest.examples.simpleled.SimpleLED` class might prove useful as an example of a simple implementation of the API.


Tested Implementations
----------------------

This version is written for MicroPython 3.4, and has been tested on:

  * Raspberry Pi Pico W

Licence
-------

This module, and all included code, is made available under the terms of the MIT Licence

> Copyright 2021--2022 (c) Erik de Lange, Copyright (c) 2022--2023 David Love

> Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

> The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

> THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

### Expose the `http` module interface
from .server import RESTServer
from .response import HTTPResponse



"""
Examples of the module use, both as simple applications and as helper classes
for reuse in more complete applications

Overview
--------

This is a collection of examples, utility classes and other scraps of code which
demonstrate the use of the library in common applications. Many of these
examples will require extension, or further development in real applications.
However they should give some idea of what those more complete applications may
look like.

Tested Implementations
----------------------

This version is written for MicroPython 3.4, and has been tested on:

  * Raspberry Pi Pico W

Licence
-------

This module, and all included code, is made available under the terms of the MIT Licence

> Copyright 2021--2022 (c) Erik de Lange, Copyright (c) 2022--2023 David Love

> Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

> The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

> THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""



"""
Formats a basic HTTP/1.1 response. The actual body of the response is largely
determined by the callers API layer: `urest.http.response.HTTPResponse` is a
utility class designed just handles to handle the raw response to the network
client. As such it should be largely invisible to the API layer, and most
consumers of the `urest.http` module _should not_ create instance of the
`urest.http.response.HTTPResponse` class directly

Standards
---------

  * For HTTP/1.1 specification see: https://www.ietf.org/rfc/rfc2616.txt
  * For MIME types see: https://www.iana.org/assignments/media-types/media-types.xhtml

Licence
-------

This module, and all included code, is made available under the terms of the MIT
Licence

> Copyright 2022 (c) Erik de Lange, Copyright (c) 2022-2023 David Love

> Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

> The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

> THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

# Import the Asynchronous IO Library, preferring the MicroPython library if
# available
try:
    import uasyncio as asyncio
except ImportError:
    import asyncio

# Import the enumerations library.
# from enum import Enum
# Unfortunately not in MicroPython yet, so we can't enforce the type
HTTPStatus = {"OK", "NOT_OK", "NOT_FOUND"}
"""
    Define the HTTP response codes in use. See the Mozilla [HTTP response status
    codes](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status) for more
    details
"""


class HTTPResponse:

    ##
    ## Constructor
    ##

    def __init__(
        self,
        body: str = "",
        status: HTTPStatus = "OK",
        mimetype: str = None,
        close: bool = True,
        header: dict = None,
    ):
        """Create a response object, representing the raw HTTP header returned to the
        network client.

        This instance is guaranteed to be a valid _class_ on creation, and
        _should_ also be a valid HTTP response. However the caller should check the
        validity of the header before returning to the client. In particular,  responses
        returned to the client by this class _must_ be formatted according to the
        [HTTP/1.1 specification](https://www.ietf.org/rfc/rfc2616.txt) and _must_ be
        valid.

        Parameters
        ----------

        body: string
            The raw HTTP body returned to the client. This is `Empty` by default as
            the return string is usually built by the caller via the `getters` and
            `setters` of `urest.http.response.HTTPResponse`
        status: urest.http.response.HTTPStatus
            HTTP status code, which must be formed from the set
            `urest.http.response.HTTPResponse`. Arbitrary return codes are **not**
            supported by this class.
        mimetype: string
            A valid HTTP mime type. This is `Empty` by default and should be set
            once the `body` of the `urest.http.response.HTTPResponse` has been
            created.
        close: bool
            When set `True` the connection to the client will be closed by the
            `urest.http.server.RESTServer` once the `urest.http.response.HTTPResponse`
            has been sent. Otherwise, when set to `False` this will flag to the
            client that the created `urest.http.response.HTTPResponse` is part of
            a sequence to be sent over the same connection.
        header:  Dictionary
            Raw (key, value) pairs for HTTP response header fields. This allows
            setting of arbitrary fields by the caller, without extending/sub-classing
            `urest.http.response.HTTPResponse`

        """

        if status in HTTPStatus:
            self._status = status
        else:
            raise ValueError(
                "Invalid HTTP status code passed to the HTTP Response class"
            )

        if body is not None and isinstance(body, str):
            self._body = body
        else:
            self._body = ""

        self._mimetype = mimetype
        self._close = close

        if header is None:
            self._header = {}
        else:
            self._header = header

    ##
    ## Getters and Setters
    ##

    # HTTP Body

    @property
    def body(self):
        """The raw HTTP response, formatted to return to the client as the HTTP response."""

        return self._body

    @body.setter
    def body(self, new_body: str):
        if new_body is not None and isinstance(new_body, str):
            self._body = new_body
        else:
            self._body = ""

    # HTTP Status

    @property
    def status(self):
        """
        A valid `urest.http.response.HTTPResponse` representing the current
        error/status code that will be returned to the client.
        """
        return self._status

    @status.setter
    def status(self, new_status: HTTPStatus):
        if new_status in HTTPStatus:
            self._status = new_status
        else:
            raise ValueError(
                "Invalid HTTP status code passed to the HTTP Response class"
            )

    ##
    ## Functions
    ##

    async def send(self, writer: asyncio.StreamWriter):
        """Send an appropriate response to the client, based on the status code.

        This method assembles the full HTTP 1.1 header, based on the `mimetype`
        the content currently in the `body`, and the error code forming the
        `status` of the response to the client.

        .. Note::
          The the actual sending of this HTTP 1.1 header to the client
          is the responsibility of the caller. This function only assists in correctly
          forming that response

        Parameters
        ----------

        writer: `asyncio.StreamWriter`
            An asynchronous stream, representing the network response to the
            client. This is usually set-up indirectly by the caller as part of a network
            response to the client. As such is is usually just a pass-though from the
            dispatch call of the server. For an example see the dispatcher
            `urest.http.server.RESTServer.dispatch_noun`.

        Returns
        -------

        `async`

          The return type is complex, and indicates this method is expected to be run
          as a co-routine under the `asyncio` library.
        """

        # **NOTE**: This implementation should be in "match/case", but MicroPython
        #       doesn't have a 3.10 release yet. When it does, this
        #       implementation should be updated

        if "OK" in self._status:
            # First tell the client we accepted the request
            writer.write(f"HTTP/1.1 200 OK\n")

            # Then we try to assemble the body
            if self._mimetype is not None:
                writer.write(f"Content-Type: {self.mimetype}\n")

        elif "NOT_OK" in self._status:
            # Tell the client we think we can route it: but the request
            # makes no sense
            writer.write(f"HTTP/1.1 400 Bad Request\n")

        elif "NOT_FOUND" in self._status:
            # Tell the client we can't route their request
            writer.write(f"HTTP/1.1 404 Not Found\n")

        else:
            # This _really_ shouldn't be here. Assume an internal error
            writer.write(f"HTTP/1.1 500 Internal Server Error\n")

        # Send the body length
        writer.write(f"Content-Length: {len(self._body)}\n")

        # Send the body content type
        writer.write(f"Content-Type: text/html\n")

        # Send any other header fields
        if len(self._header) > 0:
            for key, value in self._header.items():
                writer.write(f"{key}: {value}\n")

        # Send the HTTP connection state
        if self._close:
            writer.write("Connection: close\n")
        else:
            writer.write("Connection: keep-alive\n")

        # Send the body itself...
        writer.write(f"\n{self._body}\n")

        # ... and ensure that it gets back to the client
        await writer.drain()
